/*
 * etcp.c (C) 1995 Darren Reed
 *
 * This was written to test what size TCP fragments would get through
 * various TCP/IP packet filters, as used in IP firewalls.  In certain
 * conditions, enough of the TCP header is missing for unpredictable
 * results unless the filter is aware that this can happen.
 *
 * The author provides this program as-is, with no gaurantee for its
 * suitability for any specific purpose.  The author takes no responsibility
 * for the misuse/abuse of this program and provides it for the sole purpose
 * of testing packet filter policies.  This file maybe distributed freely
 * providing it is not modified and that this notice remains in tact.
 *
 * This was written and tested (successfully) on SunOS 4.1.x.
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#ifndef	linux
#include <netinet/ip_var.h>
#include <netinet/tcpip.h>
#endif
#include "ip_compat.h"
#ifdef	linux
#include <linux/sockios.h>
#include "tcpip.h"
#endif

extern char *optarg;
extern int optind;

char options[68];


usage(prog)
char *prog;
{
	fprintf(stderr, "Usage: %s [options] dest flags\n\
\toptions:\n\
\t\t[-d device]\n\
\t\t[-f fragflags]\n\
\t\t[-g gateway]\n\
\t\t[-m mtu]\n\
\t\t[-s src]\n\
\t\t[-t port]\n", prog);
	exit(1);
}

main(argc, argv)
int	argc;
char	**argv;
{
	struct	tcpiphdr *ti;
	struct	in_addr	gwip;
	tcphdr_t	*tcp;
	ip_t	*ip;
	char	*name =  argv[0], host[64], *gateway = NULL, *dev = NULL;
	char	*src = NULL, *dst, c, *s;
	int	mtu = 1500, olen = 0, minport = 0, maxport = 1023;

	ip = (ip_t *)calloc(1, 65536);
	ti = (struct tcpiphdr *)ip;
	tcp = (tcphdr_t *)&ti->ti_sport;
	ip->ip_len = htons(sizeof(*ip) + sizeof(*tcp));
	ip->ip_hl = sizeof(*ip) >> 2;

	while ((c = getopt(argc, argv, "d:f:g:m:o:p:P:S:s:t:")) != -1)
		switch (c)
		{
		case 'd':
			dev = optarg;
			break;
		case 'f':
			ip->ip_off = strtol(optarg, NULL, 0);
			break;
		case 'g':
			gateway = optarg;
			break;
		case 'm':
			mtu = atoi(optarg);
			if (mtu < 28)
			    {
				fprintf(stderr, "mtu must be > 28\n");
				exit(1);
			    }
			break;
		case 'o':
			olen = optname(optarg, options);
			break;
		case 'p':
			minport = atoi(optarg);
			break;
		case 'P':
			maxport = atoi(optarg);
			break;
		case 's':
			src = optarg;
			break;
		case 'S':
			tcp->th_sport = htons(atoi(optarg));
			break;
		case 't':
			tcp->th_dport = htons(atoi(optarg));
			break;
		default:
			usage(name);
		}

	if (argc - optind < 2)
		usage(name);

	dst = argv[optind++];

	if (!src) {
		gethostname(host, sizeof(host));
		src = host;
	}

	if (resolve(src, (char *)&ip->ip_src) == -1)
	    {
		fprintf(stderr,"Cant resolve %s\n", src);
		exit(2);
	    }

	if (resolve(dst, (char *)&ip->ip_dst) == -1)
	    {
		fprintf(stderr,"Cant resolve %s\n", dst);
		exit(2);
	    }

	if (!gateway)
		gwip = ip->ip_dst;
	else if (resolve(gateway, (char *)&gwip) == -1)
	    {
		fprintf(stderr,"Cant resolve %s\n", src);
		exit(2);
	    }

	for (s = argv[optind]; c = *s; s++)
		switch(c)
		{
		case 'S' : case 's' :
			tcp->th_flags |= TH_SYN;
			break;
		case 'A' : case 'a' :
			tcp->th_flags |= TH_ACK;
			break;
		case 'F' : case 'f' :
			tcp->th_flags |= TH_FIN;
			break;
		case 'R' : case 'r' :
			tcp->th_flags |= TH_RST;
			break;
		case 'P' : case 'p' :
			tcp->th_flags |= TH_PUSH;
			break;
		case 'U' : case 'u' :
			tcp->th_flags |= TH_URG;
			break;
		}

	printf("Device:  %s\n", (dev) ? dev : "default");
	printf("Source:  %s\n", inet_ntoa(ip->ip_src));
	printf("Dest:    %s\n", inet_ntoa(ip->ip_dst));
	printf("Gateway: %s\n", inet_ntoa(gwip));
	printf("Flags:   %#x\n", tcp->th_flags);
	printf("mtu:     %d\n", mtu);

	if (olen) {
		printf("Options: %d\n", olen);
		ti = (struct tcpiphdr *)malloc(olen + ntohs(ip->ip_len));
		bcopy((char *)ip, (char *)ti, sizeof(*ip));
		ip = (ip_t *)ti;
		ip->ip_hl += (olen >> 2);
		bcopy(options, (char *)(ip + 1), olen);
		bcopy((char *)tcp, (char *)(ip + 1) + olen, sizeof(*tcp));
		tcp = (tcphdr_t *)((char *)(ip + 1) + olen);
		ip->ip_len = ntohs(htons(ip->ip_len) + olen);
	}
	if (tcp->th_dport)
		return do_socket(dev, mtu, ti, gwip);
	return scan_tcp(dev, mtu, ti, gwip, minport, maxport);
}
