/*
 * (C)opyright October 1995 Darren Reed. (from tcplog)
 *
 *   This software may be freely distributed as long as it is not altered
 * in any way and that this messagge always accompanies it.
 *
 *   The author of this software makes no garuntee about the
 * performance of this package or its suitability to fulfill any purpose.
 *
 */

#include <stdio.h>
#include <netdb.h>
#include <ctype.h>
#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/mbuf.h>
#include <sys/time.h>
#include <sys/timeb.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#if BSD < 199103
#include <sys/fcntlcom.h>
#endif
#include <sys/dir.h>
#include <net/bpf.h>

#include <net/if.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/if_ether.h>
#include <netinet/ip_var.h>
#include <netinet/udp.h>
#include <netinet/udp_var.h>
#include <netinet/tcp.h>
#include <netinet/tcpip.h>

#ifndef	lint
static	char	sbpf[] = "%W% %G%";
#endif

/*
(000) ldh      [12]
(001) jeq      #0x800	   jt 2	jf 10
(002) ldb      [23]
(003) jeq      #0x6	     jt 4	jf 10
(004) ldh      [20]
(005) jset     #0x1fff	  jt 10	jf 6
(006) ldxb     4*([14]&0xf)
(007) ldh      [x + 16]
(008) jeq      #0x1	     jt 9	jf 10
(009) ret      #68
(010) ret      #0
*/
struct	bpf_insn filter[] = {
/* 0. */	{ BPF_LD|BPF_H|BPF_ABS,		0, 0, 12 },
/* 1. */	{ BPF_JMP|BPF_JEQ,		0, 8, 0x0800 },
/* 2. */	{ BPF_LD|BPF_B|BPF_ABS,		0, 0, 23 },
/* 3. */	{ BPF_JMP|BPF_JEQ,		0, 6, 0x6 },
/* 4. */	{ BPF_LD|BPF_H|BPF_ABS,		0, 0, 20 },
/* 5. */	{ BPF_JMP|BPF_JSET,		4, 0, 0x1fff },
/* 6. */	{ BPF_LDX|BPF_B|BPF_MSH,	0, 0, 14 },
/* 7. */	{ BPF_LD|BPF_H|BPF_IND,		0, 0, 16 },
/* 8. */	{ BPF_JMP|BPF_JEQ,		0, 1, 1 },
/* 9. */	{ BPF_RET,			0, 0, 68 },
/* 10. */	{ BPF_RET,			0, 0, 0 }
};
/*
 * the code herein is dervied from libpcap.
 */
static	u_char	*buf = NULL;
static	int	bufsize = 0, timeout = 1;

void	readloop(fd, port, dst)
int 	fd, port;
struct	in_addr dst;
{
	register u_char	*bp, *cp, *bufend;
	register struct	bpf_hdr	*bh;
	register int	cc;
	time_t	in = time(NULL);

	while ((cc = read(fd, buf, bufsize)) >= 0) {
		if (!cc && (time(NULL) - in) > timeout)
			return;
		bp = buf;
		bufend = buf + cc;
		/*
		 * loop through each snapshot in the chunk
		 */
		while (bp < bufend) {
			bh = (struct bpf_hdr *)bp;
			cp = bp + bh->bh_hdrlen;
			if (ack_recv(cp, port, dst))
				return;

			bp += BPF_WORDALIGN(bh->bh_caplen + bh->bh_hdrlen);
		}
	}
	perror("read");
	exit(-1);
}

int	initdevice(device, sport, tout)
char	*device;
int	sport, tout;
{
	struct	bpf_program prog;
	struct	bpf_version bv;
	struct	timeval to;
	struct	ifreq ifr;
	char	bpfname[16];
	int	fd, i;

	for (i = 0; i < 16; i++)
	    {
		(void) sprintf(bpfname, "/dev/bpf%d", i);
		if ((fd = open(bpfname, O_RDWR)) >= 0)
			break;
	    }
	if (i == 16)
	    {
		fprintf(stderr, "no bpf devices available as /dev/bpfxx\n");
		return -1;
	    }

	if (ioctl(fd, BIOCVERSION, (caddr_t)&bv) < 0)
	    {
		perror("BIOCVERSION");
		return -1;
	    }
	if (bv.bv_major != BPF_MAJOR_VERSION ||
	    bv.bv_minor < BPF_MINOR_VERSION)
	    {
		fprintf(stderr, "kernel bpf (v%d.%d) filter out of date:\n",
			bv.bv_major, bv.bv_minor);
		fprintf(stderr, "current version: %d.%d\n",
			BPF_MAJOR_VERSION, BPF_MINOR_VERSION);
		return -1;
	    }

	(void) strncpy(ifr.ifr_name, device, sizeof(ifr.ifr_name));
	if (ioctl(fd, BIOCSETIF, &ifr) == -1)
	    {
		fprintf(stderr, "%s(%d):", ifr.ifr_name, fd);
		perror("BIOCSETIF");
		exit(1);
	    }
	/*
	 * get kernel buffer size
	 */
	if (ioctl(fd, BIOCGBLEN, &bufsize) == -1)
	    {
		perror("BIOCSBLEN");
		exit(-1);
	    }
	buf = (u_char*)malloc(bufsize);
	/*
	 * set the timeout
	 */
	timeout = tout;
	to.tv_sec = 1;
	to.tv_usec = 0;
	if (ioctl(fd, BIOCSRTIMEOUT, (caddr_t)&to) == -1)
	    {
		perror("BIOCSRTIMEOUT");
		exit(-1);
	    }

	filter[8].k = htons(sport);
	prog.bf_len = sizeof(filter) / sizeof(struct bpf_insn);
	prog.bf_insns = filter;
	if (ioctl(fd, BIOCSETF, (caddr_t)&prog) == -1)
	    {
		perror("BIOCSETF");
		exit(-1);
	    }
	(void) ioctl(fd, BIOCFLUSH, 0);
	return fd;
}


int	ack_recv(ep, port, dst)
char	*ep;
int	port;
struct	in_addr dst;
{
	struct	tcpiphdr	tip;

	bcopy(ep + 14, &tip, sizeof(tip));
	if ((tip.ti_sport != port) || (dst.s_addr != tip.ti_dst.s_addr))
		return 0;
	printf("dst %s ", inet_ntoa(tip.ti_src));
	printf("src %s dport %d sport %d", inet_ntoa(tip.ti_dst),
		ntohs(tip.ti_sport), ntohs(tip.ti_dport));
	putchar(((tip.ti_flags & TH_SYN) || tip.ti_win) ? '\n' : '\r');
	return 1;
}


/*
 * output an IP packet onto a fd opened for /dev/bpf
 */
int	sendip(fd, pkt, len)
int	fd, len;
char	*pkt;
{			
	if (write(fd, pkt, len) == -1)
	    {
		perror("send");
		return -1;
	    }

	return len;
}
