/*
 * (C)opyright October 1992 Darren Reed. (from tcplog)
 *
 *   This software may be freely distributed as long as it is not altered
 * in any way and that this messagge always accompanies it.
 *
 *   The author of this software makes no garuntee about the
 * performance of this package or its suitability to fulfill any purpose.
 *
 */

#include <stdio.h>
#include <netdb.h>
#include <ctype.h>
#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/timeb.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/dir.h>
#include <linux/netdevice.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include "ip_compat.h"
#include "tcpip.h"

#ifndef	lint
static	char	sccsid[] = "%W% %G%";
#endif

#define	CHUNKSIZE	8192
#define BUFSPACE	(4*CHUNKSIZE)

/*
 * Be careful to only include those defined in the flags option for the
 * interface are included in the header size.
 */

static	int	timeout;
static	char	*eth_dev = NULL;

void	readloop(fd, port, dst)
int 	fd, port;
struct	in_addr dst;
{
	static	u_char	buf[BUFSPACE];
	struct	sockaddr dest;
	register u_char	*bp = buf;
	register int	cc;
	int	dlen;
	time_t	now = time(NULL);

	do {
		fflush(stdout);
		dlen = sizeof(dest);
		bzero((char *)&dest, dlen);
		cc = recvfrom(fd, buf, BUFSPACE, 0, &dest, &dlen);
		if (!cc)
			if ((time(NULL) - now) > timeout)
				return;
			else
				continue;

		if (bp[12] != 0x8 || bp[13] != 0)
			continue;	/* not ip */

		/*
		 * get rid of non-tcp or fragmented packets here.
		 */
		if (cc >= sizeof(struct tcpiphdr))
		    {
			if ((bp[14+9] != IPPROTO_TCP) ||
			    (bp[14+6] & 0x1f) || (bp[14+6] & 0xff))
				continue;
			if (ack_recv(bp + 14, port, dst))
				return;
		    }
	} while (cc >= 0);
	perror("read");
	exit(-1);
}

int	initdevice(dev, sport, tout)
char	*dev;
int	sport, tout;
{
	int fd;

	eth_dev = strdup(dev);
	if ((fd = socket(AF_INET, SOCK_PACKET, htons(ETHERTYPE_IP))) == -1)
	    {
		perror("socket(SOCK_PACKET)");
		exit(-1);
	    }

	return fd;
}


int	ack_recv(bp, port, dst)
char	*bp;
int	port;
struct	in_addr dst;
{
	ip_t	ip;
	tcphdr_t	tcp;

	bcopy(bp, (char *)&ip, sizeof(ip));
	bcopy(bp + (ip.ip_hl << 2), (char *)&tcp, sizeof(tcp));
	if (tcp.th_sport != htons(port) || dst.s_addr != ip.ip_dst.s_addr)
		return 0;
	printf("dst %s ", inet_ntoa(ip.ip_src));
	printf("src %s dport %d sport %d", inet_ntoa(ip.ip_dst),
		ntohs(tcp.th_sport), ntohs(tcp.th_dport));
	putchar(((tcp.th_flags & TH_SYN) || tcp.th_win) ? '\n' : '\r');
	fflush(stdout);
	sleep(1);
	return 1;
}


/*
 * output an IP packet onto a fd opened for /dev/nit
 */
int	sendip(fd, pkt, len)
int	fd, len;
char	*pkt;
{
	struct	sockaddr	s;
	struct	ifreq	ifr;

	strncpy(ifr.ifr_name, eth_dev, sizeof(ifr.ifr_name));
	if (ioctl(fd, SIOCGIFHWADDR, &ifr) == -1)
	    {
		perror("SIOCGIFHWADDR");
		return -1;
	    }
	bcopy(ifr.ifr_hwaddr.sa_data, pkt + 6, 6);
	s.sa_family = ETHERTYPE_IP;
	strncpy(s.sa_data, eth_dev, sizeof(s.sa_data));

	if (sendto(fd, pkt, len, 0, &s, sizeof(s)) == -1)
	    {
		perror("send");
		return -1;
	    }

	return len;
}
